package com.fjhx.flow.service.flow.impl;

import cn.hutool.core.util.StrUtil;
import cn.hutool.extra.spring.SpringUtil;
import com.baomidou.dynamic.datasource.toolkit.DynamicDataSourceContextHolder;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fjhx.flow.core.FlowBean;
import com.fjhx.flow.core.FlowDelegate;
import com.fjhx.flow.core.FlowThreadLocalUtil;
import com.fjhx.flow.entity.flow.dto.InitiateDto;
import com.fjhx.flow.entity.flow.po.FlowDefinition;
import com.fjhx.flow.entity.flow.po.FlowDefinitionNode;
import com.fjhx.flow.entity.flow.po.FlowExample;
import com.fjhx.flow.entity.flow.po.FlowExampleDetail;
import com.fjhx.flow.enums.FlowStatus;
import com.fjhx.flow.enums.HandleType;
import com.fjhx.flow.enums.NodeTypeEnum;
import com.fjhx.flow.service.flow.*;
import com.googlecode.aviator.AviatorEvaluator;
import com.googlecode.aviator.Expression;
import com.ruoyi.common.constant.DatasourceConstant;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.exception.ServiceException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Service
public class FlowProcessServiceImpl implements FlowProcessService {

    @Autowired
    private FlowDefinitionService flowDefinitionService;

    @Autowired
    private FlowDefinitionNodeService flowDefinitionNodeService;

    @Autowired
    private FlowExampleService flowExampleService;

    @Autowired
    private FlowExampleDetailService flowExampleDetailService;

    @Autowired
    private PlatformTransactionManager platformTransactionManager;

    @Autowired
    private TransactionDefinition transactionDefinition;

    @Override
    public void initiate(InitiateDto dto) {

        // 找到代码定义的流程bean
        Class<? extends FlowDelegate> flowDelegateCls = FlowBean.getBean(dto.getFlowKey());

        // 实例化bean
        FlowDelegate flowDelegate;

        flowDelegate = SpringUtil.getBean(flowDelegateCls);
        if (flowDelegate == null) {
            throw new ServiceException("没有把流程bean注册到spring中");
        }

        // 查找流程定义
        FlowDefinition flowDefinition = flowDefinitionService.getOne(q -> q
                .eq(FlowDefinition::getFlowKey, dto.getFlowKey())
                .eq(FlowDefinition::getCurrentVersion, StatusConstant.YES));

        if (flowDefinition == null) {
            throw new ServiceException("没有找到可用流程");
        }

        // 初始化模板参数
        Map<String, Object> templateMap = flowDelegate.initTemplateMap();
        FlowThreadLocalUtil.setTemplateData(templateMap);
        FlowThreadLocalUtil.setStartData(dto.getData());
        FlowThreadLocalUtil.setCurrentData(dto.getData());

        // 流程节点列表
        List<FlowDefinitionNode> flowDefinitionNodeList = flowDefinitionNodeService
                .list(q -> q.eq(FlowDefinitionNode::getFlowDefinitionId, flowDefinition.getId()));

        // 父级节点map
        Map<Long, List<FlowDefinitionNode>> parentNodeMap = flowDefinitionNodeList.stream()
                .collect(Collectors.groupingBy(FlowDefinitionNode::getParentId));

        // 寻找开始节点
        FlowDefinitionNode startNode = flowDefinitionNodeList.stream()
                .filter(item -> NodeTypeEnum.START.getKey().equals(item.getNodeType()))
                .findFirst().orElse(null);

        if (startNode == null) {
            throw new ServiceException("没有找到开始节点");
        }

        // 寻找下一节点
        FlowDefinitionNode nextUserNode = getNextUserNode(startNode, parentNodeMap);

        // 开启一个事务
        TransactionStatus transaction = platformTransactionManager.getTransaction(transactionDefinition);

        try {
            // 生成流程id
            long flowId = IdWorker.getId();
            FlowThreadLocalUtil.setFlowId(flowId);

            // 切换到主库执行流程开始节点逻辑
            // DynamicDataSourceContextHolder.push(DatasourceConstant.MASTER_NAME);

            FlowExample flowExample = new FlowExample();
            List<FlowExampleDetail> flowExampleDetailList = new ArrayList<>();

            // 执行开始流程方法
            Long businessId = flowDelegate.start(flowId, dto.getData());
            FlowThreadLocalUtil.setBusinessId(businessId);

            FlowExampleDetail startExampleDetail = new FlowExampleDetail();
            startExampleDetail.setFlowExampleId(flowId);
            startExampleDetail.setFlowDefinitionNodeId(startNode.getId());
            startExampleDetail.setFlowDefinitionNodeType(startNode.getNodeType());
            startExampleDetail.setHandleType(HandleType.NEXT.getKey());
            flowExampleDetailList.add(startExampleDetail);

            // 如果下个节点是结束节点
            if (NodeTypeEnum.END.equals(NodeTypeEnum.getEnum(nextUserNode.getNodeType()))) {
                // 走非默认方法
                String handlingMethod = nextUserNode.getHandlingMethod();
                if (StrUtil.isNotBlank(handlingMethod)) {
                    Method method = flowDelegateCls.getMethod(handlingMethod);
                    method.invoke(flowDelegate);
                } else {
                    // 执行流程结束方法
                    flowDelegate.end();
                }

                // 流程已通过
                flowExample.setStatus(FlowStatus.HAVE_PASSED.getKey());

                // 结束流程
                FlowExampleDetail endExampleDetail = new FlowExampleDetail();
                startExampleDetail.setFlowExampleId(flowId);
                startExampleDetail.setFlowDefinitionNodeId(nextUserNode.getId());
                startExampleDetail.setFlowDefinitionNodeType(nextUserNode.getNodeType());
                startExampleDetail.setHandleType(HandleType.OVER.getKey());
                flowExampleDetailList.add(endExampleDetail);
            } else {
                // 流程进行中
                flowExample.setStatus(FlowStatus.IN_PROGRESS.getKey());
            }

            // 切换到从库执行流程信息保存
            DynamicDataSourceContextHolder.push(DatasourceConstant.BASE);

            flowExample.setTitle(StrUtil.format(flowDefinition.getTitleTemplate(), templateMap, false));
            flowExample.setDefinitionId(flowDefinition.getId());
            flowExample.setDefintionNodeId(nextUserNode.getId());
            flowExample.setBusinessId(businessId);
            flowExample.setStartData(dto.getData().toJSONString());
            flowExampleService.save(flowExample);
            flowExampleDetailService.saveBatch(flowExampleDetailList);

            // 提交事务
            platformTransactionManager.commit(transaction);
        } catch (Exception e) {
            // 回滚事务
            platformTransactionManager.rollback(transaction);
        }

    }

    /**
     * 查找下一个用户执行节点
     *
     * @param currentNode   当前节点
     * @param parentNodeMap 父节点map
     * @return 用户节点
     */
    private FlowDefinitionNode getNextUserNode(FlowDefinitionNode currentNode, Map<Long, List<FlowDefinitionNode>> parentNodeMap) {

        // 查找下个节点
        List<FlowDefinitionNode> nextNodeList = parentNodeMap.get(currentNode.getId());

        // 如果不为分支，返回此节点
        if (!NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(currentNode.getNodeType()))) {

            if (nextNodeList.size() != 1) {
                throw new ServiceException("流程定义错误");
            }

            return nextNodeList.get(0);
        }

        // 找到分支跳转节点
        FlowDefinitionNode defaultDefinitionNode = null;
        Map<String, Object> templateData = FlowThreadLocalUtil.getTemplateData();
        for (FlowDefinitionNode flowDefinitionNode : nextNodeList) {
            String jumpCondition = flowDefinitionNode.getJumpCondition();
            if (StrUtil.isBlank(jumpCondition)) {
                defaultDefinitionNode = flowDefinitionNode;
            }
            if (expressionResult(templateData, jumpCondition)) {
                return flowDefinitionNode;
            }
        }

        if (defaultDefinitionNode == null) {
            throw new ServiceException("未识别到分支跳转节点");
        }

        return defaultDefinitionNode;
    }

    /**
     * 校验el表达式
     */
    public static boolean expressionResult(Map<String, Object> map, String expression) {
        Expression exp = AviatorEvaluator.compile(expression);
        Object execute = exp.execute(map);
        return Boolean.parseBoolean(String.valueOf(execute));
    }

    // public static void main(String[] args) {
    //     // HashMap<Object, Object> map1 = new HashMap<>();
    //     // map1.put("test2", "2");
    //     //
    //     HashMap<String, Object> map2 = new HashMap<>();
    //
    //     map2.put("admin3", "dsaasdsad");
    //     map2.put("admin1", 2);
    //
    //     System.out.println(expressionResult(map2, "admin1 > 2"));
    // }

}
